# -*- coding: utf-8 -*-
#
# django-codenerix
#
# Codenerix GNU
#
# Project URL : http://www.codenerix.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

try:
    from subprocess import getstatusoutput
    pythoncmd = "python3"
except:
    from commands import getstatusoutput
    pythoncmd = "python"

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from codenerix.lib.debugger import Debugger


class Command(BaseCommand, Debugger):

    # Show this when the user types help
    help = "Used to build all locales details"
    
    def add_arguments(self, parser):
        # Set witch mode to use
        parser.add_argument('--mode',
                            dest='mode',
                            default="suexec",
                            help="Mode used in the environment 'suexec', 'apache' or 'wwwdata'") 
        
        # Named (optional) arguments
        parser.add_argument('--noauto',
                            action='store_true',
                            dest='noauto',
                            default=False,
                            help='Tells the command not to find automatic solution for problems')
        
        # Named (optional) arguments
        parser.add_argument('--noguess',
                            action='store_true',
                            dest='noguess',
                            default=False,
                            help='Disable guessing user environment')

        # Named (optional) arguments
        parser.add_argument('--clean',
                            action='store_true',
                            dest='clean',
                            default=False,
                            help='Do a full clean by deleting everything before building translations')

        # Named (optional) arguments
        parser.add_argument('--compile',
                            action='store_true',
                            dest='compile',
                            default=False,
                            help='Compile .po files')

    def handle(self, *args, **options):
        # Check arguments
        mode = options['mode']
        if mode not in ['wwwdata', 'apache', 'suexec']:
            self.print_help('', '')
            raise CommandError("Allowed modes are suexec, apache or wwwdata ('apache' and 'wwwdata' option will force permissions to apache or www-data user, suexec won't touch permissions)")
        
        # Autoconfigure Debugger
        self.set_name("CODENERIX")
        self.set_debug()
        
        # Get environment
        appname = settings.ROOT_URLCONF.split(".")[0]
        basedir = settings.BASE_DIR
        appdir = os.path.abspath("{}/{}".format(basedir, appname))
        noauto = options['noauto']

        # Check user selection
        if not options['noguess']:
            cmd = "find {} -name 'locale' -exec ls -lR {{}} \; | grep www-data".format(appdir)
            status, output = getstatusoutput(cmd)
            if status:
                guess = "suexec"
            else:
                guess = "wwwdata"
            if guess != mode:
                self.print_help('', '')
                raise CommandError("You have selected mode '{}' but I believe you are using '{}', if sure use --noguess".format(mode, guess))
        
        # Show header
        self.debug("Creating locales for {}".format(appname), color='blue')
        if noauto:
            self.debug("Autoconfig mode is OFF", color='yellow')
        
        # Get list of apps
        apps = []
        length = len(appname)
        for app in settings.INSTALLED_APPS:
            if app[0:length] == appname:
                apps.append(app[length + 1:])
        
        # Check for locale folders
        ERROR = False
        for app in [''] + apps:
            testpath = os.path.abspath("{}/{}/locale".format(appdir, app).replace("//", "/"))
            if not os.path.exists(testpath):
                if noauto:
                    ERROR = True
                    self.debug("'locale' folder missing at {}/".format(testpath), color='yellow')
                else:
                    self.debug("'locale' folder missing, creating {}/".format(testpath), color='purple')
                    os.mkdir(testpath)
        
        if ERROR:
            raise CommandError("Some error has happened, can not keep going! (avoid using --noauto to let CODENERIX find a solution)")
        
        # Check execution mode
        sudo = ''
        if mode == 'apache' or mode == 'wwwdata':
            status, output = getstatusoutput("whoami")
            if status:
                # Error in command
                self.error("Error while executing 'whoami' command")
                raise CommandError(output)
            elif output == 'www-data':
                # Detected we are already www-data user, so keep going
                self.debug("Detected we are 'www-data' user", color='purple')
            else:
                # No permissions try to become root using sudo
                status, output = getstatusoutput("sudo whoami")
                if status:
                    # Error in command
                    self.error("Error while executing 'sudo whoami' command")
                    raise CommandError(output)
                elif output == 'root':
                    # Detected sudo is working and we can become root
                    self.debug("Detected we can become 'root' user", color='purple')
                    sudo = 'sudo '
                else:
                    # No permissions
                    raise CommandError("You requested 'apache' or 'wwwdata' execution mode but you are not www-data and we can not become root")
        
        # Delete locale folders if requested
        if options['clean']:
            
            # Ask the user if would really like to remove all locale folders
            key = ''
            while key not in ['n', 'y']:
                self.debug("All 'locale' folders are going to be removed, are you sure? (y|n) ", tail=False, color="red")
                try:
                    key = raw_input().lower()
                except NameError:
                    key = input().lower()
                self.debug("", header=False)
            
            # Remove all locale folders
            if key == 'y':
                self.debug("Removing locale folders...", color="red")
                for app in [''] + apps:
                    testpath = os.path.abspath("{}/{}/locale".format(appdir, app).replace("//", "/"))
                    if os.path.exists(testpath):
                        self.debug("    > Removing {}".format(testpath), color="red")
                        cmd = "{}rm -R {}/*".format(sudo, testpath)
                        status, output = getstatusoutput(cmd)
                        if status:
                            raise CommandError(output)
            else:
                raise CommandError("You requested to clean all 'locale' folders but you answered NO to the previous question, can not keep going!")
        
        # Ready to go
        if not options['compile']:
            for (code, name) in settings.LANGUAGES:
                self.debug("Processing translations for {}...".format(name), color='cyan')
                cmd = "{}{}/manage.py makemessages -v0 --symlinks --ignore env -l {}".format(sudo, basedir, code)  # --exclude is not working in some computers
                status, output = getstatusoutput(cmd)
                if status:
                    raise CommandError(output)
                
                cmd = "{}{}/manage.py makemessages -v0 --symlinks --ignore env -d djangojs -l {}".format(sudo, basedir, code)  # --exclude is not working in some computers
                status, output = getstatusoutput(cmd)
                if status:
                    raise CommandError(output)

        # Set permissions for locale folders
        if sudo:
            self.debug("Setting permissions...", color='cyan')
            if mode == 'apache':
                user = 'apache'
            elif mode == 'wwwdata':
                user = 'www-data'
            else:
                raise CommandError("Wrong mode for sudo '{}'".format(mode))
            for app in [''] + apps:
                testpath = os.path.abspath("{}/{}/locale".format(appdir, app).replace("//", "/"))
                cmd = "sudo chown {}.{} {}/ -R".format(user, user, testpath)
                status, output = getstatusoutput(cmd)
                if status:
                    raise CommandError(output)




# -*- coding: utf-8 -*-
#
# django-codenerix
#
# Codenerix GNU
#
# Project URL : http://www.codenerix.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# System
import os
from datetime import datetime
import random
from dateutil.tz import tzutc
import zipfile
import io
import json
from unidecode import unidecode
from uuid import UUID
import time
import decimal
import importlib


# Django
from django.db.models import Q
from django.utils.encoding import smart_str
from django.utils.translation import gettext_lazy as _
from django.template import TemplateDoesNotExist
from django.template.loader import get_template as django_get_template
from django.shortcuts import render
from django.conf import settings
from django.views.generic.base import View
from django.core.cache import cache
from django.utils import dateparse
from django.http import HttpResponseRedirect
from django.utils.http import urlsafe_base64_encode
from django.urls import reverse_lazy


def epochdate(timestamp):
    '''
    Convet an epoch date to a tuple in format ("yyyy-mm-dd","hh:mm:ss")
    Example: "1023456427" -> ("2002-06-07","15:27:07")

    Parameters:
    - `timestamp`: date in epoch format
    '''

    dt = datetime.fromtimestamp(float(timestamp)).timetuple()
    fecha = "{0:d}-{1:02d}-{2:02d}".format(dt.tm_year, dt.tm_mon, dt.tm_mday)
    hora = "{0:02d}:{1:02d}:{2:02d}".format(dt.tm_hour, dt.tm_min, dt.tm_sec)
    return (fecha, hora)


def date2string(dtime, formt, default):
    try:
        list_type = [str, unicode, ]
    except NameError:
        list_type = [str, ]

    if dtime:
        if ('year' in dir(dtime)) and (dtime.year < 1900):
            return default
        elif type(dtime) in list_type:
            return dtime
        else:
            return dtime.strftime(formt)
    else:
        return default


def daterange_filter(value, variable):
    start = dateparse.parse_datetime(value['startDate'])
    end = dateparse.parse_datetime(value['endDate'])
    result = {'{0}__gte'.format(variable): start, '{0}__lte'.format(variable): end}
    return result


def timezone_serialize(d, prettify=False):
    if d is None:
        result = None
    else:
        if d.tzinfo:
            d = d.astimezone(tzutc()).replace(tzinfo=None)
        if prettify:
            result = d.strftime(settings.DATETIME_INPUT_FORMATS[0])
        else:
            result = d.isoformat() + "Z"
    return result


def timezone_deserialize(string):
    if string is None:
        result = None
    else:
        result = dateparse.parse_datetime(string)
    return result


def zeropad(number, width):
    return '{number:0{width}d}'.format(width=width, number=number)


def monthname(value):
    if value == 1:
        return _("January")
    elif value == 2:
        return _("February")
    elif value == 3:
        return _("March")
    elif value == 4:
        return _("April")
    elif value == 5:
        return _("May")
    elif value == 6:
        return _("June")
    elif value == 7:
        return _("July")
    elif value == 8:
        return _("August")
    elif value == 9:
        return _("September")
    elif value == 10:
        return _("Octuber")
    elif value == 11:
        return _("November")
    elif value == 12:
        return _("December")
    else:
        return None


# Name unify
def nameunify(name, url=False):
    # Make unicode
    name = smart_str(name)

    # Get it on lower
    namelow = unidecode(name).lower().strip(' \t\n\r')

    # Define valid characters
    allowed_characters = 'abcdefghijklmnopqrstuvwxyz0123456789_-'
    if url:
        namelow = namelow.replace(' ', "-")
    else:
        namelow = namelow.replace("-", "").replace(' ', '_')

    # Replace all unknown characters
    result = ''
    for c in namelow:
        if c in allowed_characters:
            result += c

    if result[-1] in ("-", "_"):
        result = nameunify(result[0:-1], url)
    # Return the cleaned result
    return result


def get_profile(user):
    # Check if it has admin rights admin
    try:
        is_admin = bool(user.is_superuser or user.groups.get(name='Admins'))
    except Exception:
        is_admin = False

    if is_admin:
        # Administrator use root
        profile = "admin"
    else:
        profile = None

    # Return profile
    return profile


def get_profiled_paths(path, user, lang, extension):
    paths = []
    if user:
        # Check if it has admin rights admin
        profile = get_profile(user)

        # Define the base_path to use
        paths.append("{0}.{1}".format(user.username, lang))
        paths.append(user.username)
        if profile:
            paths.append("{0}.{1}".format(profile, lang))
            paths.append(profile)
        if profile != 'admin':
            paths.append("user.{0}".format(lang))
            paths.append('user')

    # Add an empty path
    paths.append(lang)
    paths.append("")

    # Split path name
    pathsp = path.split(".")
    if pathsp[-1] == extension:
        basepath = ".".join(pathsp[:-1])
    else:
        basepath = path

    # Return paths and basepath
    return (paths, basepath)


def get_template(template, user, lang, extension='html', raise_error=True):
    # Get profiled paths
    (templates, templatepath) = get_profiled_paths(template, user, lang, extension)

    # Check templates
    test = []
    found = None
    for temp in templates:
        if len(temp):
            addon = ".%s" % (temp)
        else:
            addon = ""
        target = "%s%s.%s" % (templatepath, addon, extension)
        test.append(target)
        try:
            django_get_template(target)
            found = target
            break
        except TemplateDoesNotExist:
            pass

    # Return target template
    if found:
        return found
    else:
        if raise_error:
            raise IOError("I couldn't find a valid template, I have tried: {}".format(test))
        else:
            return test


def get_static(desired, user, lang, default, extension='html', relative=False):
    # Get profiled paths
    (paths, basepath) = get_profiled_paths(desired, user, lang, extension)

    # Check paths
    found = None
    for temp in paths:
        if len(temp):
            addon = ".%s" % (temp)
        else:
            addon = ""
        target = "%s%s.%s" % (basepath, addon, extension)
        if hasattr(settings, "STATIC_ROOT") and settings.STATIC_ROOT and os.path.exists(os.path.join(settings.STATIC_ROOT, target)):
            if relative:
                found = os.path.join(settings.STATIC_URL, target)
            else:
                found = target
            break

    # Return target template
    if found:
        # Return the one we found
        return found
    else:
        # We didn't find any, return default one
        return default


def direct_to_template(request, template, lang):
    template = get_template(template, request.user)
    return render(request, template, lang)


def model_inspect(obj):
        '''
        Analize itself looking for special information, right now it returns:
        - Application name
        - Model name
        '''
        # Prepare the information object
        info = {}
        if hasattr(obj, '_meta'):
            info['verbose_name'] = getattr(obj._meta, 'verbose_name', None)
        else:
            info['verbose_name'] = None

        # Get info from the object
        if hasattr(obj, 'model') and obj.model:
            model = obj.model
        else:
            model = obj.__class__

        namesp = str(model)
        namesp = namesp.replace("<class ", "").replace(">", "").replace("'", "").split(".")

        # Remember information
        info['appname'] = namesp[-3]
        info['modelname'] = namesp[-1]
        info['model'] = model

        # Return the info
        return info


def upload_path(instance, filename):
    '''
    This method is created to return the path to upload files. This path must be
    different from any other to avoid problems.
    '''
    path_separator = "/"
    date_separator = "-"
    ext_separator = "."
    empty_string = ""
    # get the model name
    model_name = model_inspect(instance)['modelname']

    # get the string date
    date = datetime.now().strftime("%Y-%m-%d").split(date_separator)
    curr_day = date[2]
    curr_month = date[1]
    curr_year = date[0]

    split_filename = filename.split(ext_separator)
    filename = empty_string.join(split_filename[:-1])
    file_ext = split_filename[-1]

    new_filename = empty_string.join([filename, str(random.random()).split(ext_separator)[1]])
    new_filename = ext_separator.join([new_filename, file_ext])
    string_path = path_separator.join([model_name, curr_year, curr_month, curr_day, new_filename])
    # the path is built using the current date and the modelname
    return string_path


def get_class(func):
    if not getattr(func, '__closure__', None):
        return

    for closure in func.__closure__:
        contents = closure.cell_contents

        if not contents:
            continue

        if getattr(contents, '__bases__', None) and issubclass(contents, View):
            return contents

        result = get_class(contents)
        if result:
            return result


def get_client_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


def clean_memcache_item(key, item):
    # clean item from key
    result = cache.get(key)
    if result is not None and item in result:
        result.pop(item)
        cache.set(key, result)


class CodenerixEncoder(object):

    codenerix_numeric_dic = {
        # Basic dicts
        'num': '0123456789',
        'alpha': 'ABCDEFGHIJKLMNOPQRSTUVWXYZ',
        'alphanc': 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ',
        'alphanum': '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ',
        'alphanumnc': '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ',
        # Original dicts
        'hex36': 'ENWR6MV71JOQHADUGFCYZ25X3B4L0KPTSI98',
        'hex16': '54BEF80D1C96A732',
        'hexz17': 'Z0123456789ABCDEF',
    }

    def list_encoders(self):
        return self.codenerix_numeric_dic.keys()

    def numeric_encode(self, number, dic='hex36', length=None, cfill=None):

        # Get predefined dictionary
        if dic in self.codenerix_numeric_dic:
            dic = self.codenerix_numeric_dic[dic]

        # Integrity check to dic
        nr = ""
        for c in dic:
            if c not in nr:
                nr += c
            else:
                raise IOError(_("ERROR: dic has repeated elements"))

        # If no cfill
        if cfill is None:
            cfill = dic[0]

        # Find lenght
        ldic = len(dic)

        # Initialize
        string = ""
        div = ldic + 1

        # Process number
        while div >= ldic:
            div = int(number / ldic)
            mod = number % ldic
            string += dic[mod]
            number = div

        # If something left behind
        if div:
            string += dic[div]

        # Fill the string if requested
        if length:
            string += cfill * (length - len(string))

        # Return the reverse string
        return string[::-1]

    def numeric_decode(self, string, dic='hex36'):

        # Get predefined dictionary
        if dic in self.codenerix_numeric_dic:
            dic = self.codenerix_numeric_dic[dic]

        # For each character in the string
        first = True
        number = 0
        for c in string:

            # Not the first loop
            if not first:
                number *= len(dic)
            else:
                first = False

            # Attach the character
            number += dic.index(c)

        # Return the final resulting number
        return number


class InMemoryZip(object):
    '''
    # Compress
    imz = InMemoryZip()
    imz.append("info.dat", data).append("test.txt","asdfasfdsaf")
    datazip = imz.read()

    # Get FILE pointer and read() from it
    FILE = imz.get_fp()
    FILE.read()

    # Uncompress
    imz = InMemoryZip(datazip)$
    info_unzipped = imz.get("info.dat")
    '''

    def __init__(self, data=None):
        # Create the in-memory file-like object
        self.in_memory_zip = io.BytesIO()
        if data:
            self.in_memory_zip.write(data)
            self.in_memory_zip.seek(0)

    def get_fp(self):
        self.in_memory_zip.seek(0)
        return self.in_memory_zip

    def append(self, filename_in_zip, file_contents):
        '''
        Appends a file with name filename_in_zip and contents of
        file_contents to the in-memory zip.
        '''
        # Set the file pointer to the end of the file
        self.in_memory_zip.seek(-1, io.SEEK_END)

        # Get a handle to the in-memory zip in append mode
        zf = zipfile.ZipFile(self.in_memory_zip, "a", zipfile.ZIP_DEFLATED, False)

        # Write the file to the in-memory zip
        zf.writestr(filename_in_zip, file_contents)

        # Mark the files as having been created on Windows so that
        # Unix permissions are not inferred as 0000
        for zfile in zf.filelist:
            zfile.create_system = 0

        # Close the ZipFile
        zf.close()

        # Rewind the file
        self.in_memory_zip.seek(0)

        return self

    def get(self, filename):
        self.in_memory_zip.seek(0)
        zf = zipfile.ZipFile(self.in_memory_zip, "r", zipfile.ZIP_DEFLATED, False)
        fp = zf.open(filename)
        data = fp.read()
        # Close the ZipFile
        zf.close()
        # Return the result
        return data

    def size(self):
        size = self.in_memory_zip.seek(-1, io.SEEK_END)
        self.in_memory_zip.seek(0)
        return size

    def read(self):
        '''Returns a string with the contents of the in-memory zip.'''
        self.in_memory_zip.seek(0)
        return self.in_memory_zip.read()

    def writetofile(self, filename):
        '''Writes the in-memory zip to a file.'''
        f = open(filename, "w")
        f.write(self.read())
        f.close()


def remove_getdisplay(field_name):
    '''
    for string 'get_FIELD_NAME_display' return 'FIELD_NAME'
    '''
    str_ini = 'get_'
    str_end = '_display'
    if str_ini == field_name[0:len(str_ini)] and str_end == field_name[(-1) * len(str_end):]:
        field_name = field_name[len(str_ini):(-1) * len(str_end)]
    return field_name


def trace_json_error(struct, path=[]):
    found = None
    kind = type(struct)
    if kind is dict:
        for key in struct:
            found = trace_json_error(struct[key], path + [key])
            if found:
                break
    elif kind is list:
        idx = 0
        for element in struct:
            found = trace_json_error(element, path + [str(idx)])
            if found:
                break
            idx += 1
    else:
        try:
            json.dumps(struct)
        except Exception:
            found = path

    # Return result
    return found


def qobject_builder_string_search(valid_fields, text, conditions='icontains'):
    fields = None
    for word in text.split(" "):
        if word and word[0] == '-':
            word = word[1:]
            neg = True
        else:
            neg = False

        temp = None
        for field in valid_fields:
            qobject = Q(**{'{}__{}'.format(field, conditions): word})
            if temp:
                temp |= qobject
            else:
                temp = qobject

        if temp:
            if neg:
                temp = ~Q(temp)

            if fields:
                fields &= temp
            else:
                fields = temp
    return fields


def form_answer(status, answer):
    # Normalize answer
    if '__pk__' not in answer:
        answer['__pk__'] = None
    if '__str__' not in answer:
        answer['__str__'] = 'OK'

    # Encode answer
    answer_encoded = urlsafe_base64_encode(str.encode(json.dumps(answer))).decode()

    # Build success URL
    success_url = reverse_lazy("CDNX_status", kwargs={'status': 'accept', 'answer': answer_encoded})

    # Return response
    return HttpResponseRedirect(success_url)


def JSONEncoder_newdefault(kind=['uuid', 'datetime', 'time', 'decimal']):
    '''
    JSON Encoder newdfeault is a wrapper capable of encoding several kinds
    Usage:
        from codenerix.helpers import JSONEncoder_newdefault
        JSONEncoder_newdefault()
    '''
    JSONEncoder_olddefault = json.JSONEncoder.default

    def JSONEncoder_wrapped(self, o):
        '''
        json.JSONEncoder.default = JSONEncoder_newdefault
        '''
        if ('uuid' in kind) and isinstance(o, UUID):
            return str(o)
        if ('datetime' in kind) and isinstance(o, datetime):
            return str(o)
        if ('time' in kind) and isinstance(o, time.struct_time):
            return datetime.fromtimestamp(time.mktime(o))
        if ('decimal' in kind) and isinstance(o, decimal.Decimal):
            return str(o)
        return JSONEncoder_olddefault(self, o)
    json.JSONEncoder.default = JSONEncoder_wrapped


def context_processors_update(context, request):
    '''
    Update context with context_processors from settings
    Usage:
        from codenerix.helpers import context_processors_update
        context_processors_update(context, self.request)
    '''
    for template in settings.TEMPLATES:
        for context_processor in template['OPTIONS']['context_processors']:
            path = context_processor.split('.')
            name = path.pop(-1)
            processor = getattr(importlib.import_module('.'.join(path)), name, None)
            if processor:
                context.update(processor(request))
    return context



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('codenerix', '0003_auto_20160817_1235'),
    ]

    operations = [
        migrations.AlterField(
            model_name='log',
            name='action_flag',
            field=models.PositiveSmallIntegerField(verbose_name='Action'),
        ),
    ]



__version__ = "4.0.5"

__authors__ = [
    "Juanmi Taboada",
    "Juan Soler Ruiz",
]

__authors_email__ = [
    "juanmi@juanmitaboada.com",
    "soleronline@gmail.com",
]


__requirements__ = {
    "all": [
        "pymongo",
        "python-dateutil",
        "django-recaptcha>=1.2.1,<1.3",
        "django-rosetta>=0.9.8",
        "jsonfield",
        "openpyxl",
        "Pillow",
        "Unidecode",
        "Django>=2.2.9,<=4.0.9",
        "ldap3",
        "django-haystack>=2.6.1",
        "pytz",
        "elasticsearch>=2.0.0,<3.0.0",
        "django-debug-toolbar==3.2.2",
        "pyotp",
    ],
    "2": [
        "html5lib==1.0b8",  # Default version 0.99999999 is broken with error 'from html5lib import treebuilders, inputstream' => 'ImportError: cannot import name inputstream' (1.0b10 also fails)
    ],
    "3": [
        "html5lib",
    ],
}



from django import forms
from django.core.validators import validate_email
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from .widgets import MultiEmailWidget


class MultiEmailField(forms.Field):
    message = _('Enter valid email addresses.')
    code = 'invalid'
    widget = MultiEmailWidget

    def to_python(self, value):
        "Normalize data to a list of strings."
        # Return None if no input was given.
        if not value:
            return []
        return [v.strip() for v in value.splitlines() if v != ""]

    def validate(self, value):
        """ Check if value consists only of valid emails. """

        # Use the parent's handling of required fields, etc.
        super(MultiEmailField, self).validate(value)
        try:
            for email in value:
                validate_email(email)
        except ValidationError:
            raise ValidationError(self.message, code=self.code)



# -*- coding: utf-8 -*-
from django import VERSION
from django.forms.forms import DeclarativeFieldsMetaclass, BaseForm
from django.forms.models import BaseModelForm
import six
from .angular_base import BaseFieldsModifierMetaclass, NgFormBaseMixin
from .angular_model import NgModelFormMixin
if VERSION[:2] >= (1, 5):
    from .angular_validation import NgFormValidationMixin
if VERSION[:2] < (1, 7):
    from .models import PatchedModelFormMetaclass as ModelFormMetaclass
else:
    from django.forms.models import ModelFormMetaclass


class NgDeclarativeFieldsMetaclass(BaseFieldsModifierMetaclass, DeclarativeFieldsMetaclass):
    pass


class NgForm(six.with_metaclass(NgDeclarativeFieldsMetaclass, NgFormBaseMixin, BaseForm)):
    """
    Convenience class to be used instead of Django's internal ``forms.Form`` when declaring
    a form to be used with AngularJS.
    """


class NgModelFormMetaclass(BaseFieldsModifierMetaclass, ModelFormMetaclass):
    pass


class NgModelForm(six.with_metaclass(NgModelFormMetaclass, NgFormBaseMixin, BaseModelForm)):
    """
    Convenience class to be used instead of Django's internal ``forms.ModelForm`` when declaring
    a model form to be used with AngularJS.
    """



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from django.forms import widgets
from django.utils.html import format_html
from django.utils.encoding import force_str
from .angular_base import NgFormBaseMixin, SafeTuple


class NgFormValidationMixin(NgFormBaseMixin):
    """
    Add this NgFormValidationMixin to every class derived from forms.Form, which shall be
    auto validated using the Angular's validation mechanism.
    """
    add_djng_error = True

    def __init__(self, *args, **kwargs):
        super(NgFormValidationMixin, self).__init__(*args, **kwargs)
        for name, field in self.fields.items():
            # add ng-model to each model field
            ng_model = self.add_prefix(name)
            field.widget.attrs.setdefault('ng-model', ng_model)

    def get_field_errors(self, bound_field):
        """
        Determine the kind of input field and create a list of potential errors which may occur
        during validation of that field. This list is returned to be displayed in '$dirty' state
        if the field does not validate for that criteria.
        """
        errors = super(NgFormValidationMixin, self).get_field_errors(bound_field)
        if bound_field.is_hidden:
            return errors
        identifier = format_html('{0}[\'{1}\']', self.form_name, self.add_prefix(bound_field.name))
        potential_errors = bound_field.field.get_potential_errors()
        errors.extend([SafeTuple((identifier, self.field_error_css_classes, '$dirty', pe[0], 'invalid', force_str(pe[1])))
                       for pe in potential_errors])
        if not isinstance(bound_field.field.widget, widgets.PasswordInput):
            # all valid fields shall display OK tick after changed into dirty state
            errors.append(SafeTuple((identifier, self.field_error_css_classes, '$dirty', '$valid', 'valid', '')))
            if bound_field.value():
                # valid bound fields shall display OK tick, even in pristine state
                errors.append(SafeTuple((identifier, self.field_error_css_classes, '$pristine', '$valid', 'valid', '')))
        return errors

    def get_widget_attrs(self, bound_field):
        attrs = super(NgFormValidationMixin, self).get_widget_attrs(bound_field)
        # transfer error state from bound field to AngularJS validation
        errors = [e for e in bound_field.errors if e[3] == '$pristine']
        if errors and self.add_djng_error:
            attrs.update({'djng-error': 'bound-field'})
        # some fields require special directives to work with AngularJS
        try:
            attrs.update(bound_field.field.widget.get_field_attrs(bound_field.field))
        except AttributeError:
            pass
        return attrs



from pyquery import PyQuery as pq

from django import forms
from django.core.exceptions import ValidationError
from django.test import SimpleTestCase, TestCase

from multi_email_field.forms import MultiEmailField as MultiEmailFormField
from multi_email_field.tests.models import TestModel
from multi_email_field.widgets import MultiEmailWidget


class MultiEmailFormTest(SimpleTestCase):
    def test__form(self):
        class TestForm(forms.Form):
            f = MultiEmailFormField()

        form = TestForm(initial={'f': ['foo@foo.fr', 'bar@bar.fr']})
        output = form.as_p()
        self.assertEqual(1, len(pq('textarea', output)))
        # The template-based widget add a line-return
        value = pq('textarea', output).text()
        self.assertEqual(
            value.lstrip(),
            'foo@foo.fr\nbar@bar.fr'
        )


class MultiEmailModelFormTest(TestCase):
    def test__model_form(self):
        class TestModelForm(forms.ModelForm):
            class Meta:
                model = TestModel
                fields = forms.ALL_FIELDS

        form = TestModelForm()
        self.assertIsInstance(form.fields['f'].widget, MultiEmailWidget)

        form = TestModelForm(data={'f': 'foo@foo.fr\nbar@bar.fr'})
        self.assertTrue(form.is_valid())
        instance = form.save()
        self.assertEqual(instance.f, ['foo@foo.fr', 'bar@bar.fr'])

        form = TestModelForm(instance=instance)
        output = form.as_p()
        self.assertEqual(2, len(pq('textarea', output)))
        # The template-based widget add a line-return
        value = pq('textarea:first', output).text()
        self.assertEqual(
            value.lstrip(),
            'foo@foo.fr\nbar@bar.fr',
        )


class MultiEmailModelTest(TestCase):
    def test_default_value(self):
        self.assertEqual(TestModel().f, [])

    def test_custom_default_value(self):
        self.assertEqual(TestModel().f_default, ["test@example.com"])

    def test__clean(self):
        # Nothing of these should raise a ValidationError
        TestModel().full_clean()
        TestModel(f="").full_clean()
        TestModel(f=None).full_clean()
        TestModel(f=False).full_clean()
        TestModel(f='foobar@foobar.fr\nbaz@baz.fr').full_clean()
        TestModel(f=['foo@foo.fr', 'bar@bar.fr']).full_clean()

    def test__string_based_value(self):
        emails = 'foobar@foobar.fr\nbaz@baz.fr'
        TestModel.objects.create(f=emails)
        values = TestModel.objects.values_list('f', flat=True)
        self.assertEqual(values[0], emails.splitlines())

        qs = TestModel.objects.all()
        self.assertTrue(qs.filter(f=emails).exists())

    def test__list_based_value(self):
        emails = ['foo@foo.fr', 'bar@bar.fr']
        TestModel.objects.create(f=emails)

        values = TestModel.objects.values_list('f', flat=True)
        self.assertEqual(values[0], emails)

    def test__empty_based_value(self):
        qs = TestModel.objects.all()

        TestModel.objects.create(f=[])
        self.assertEqual(qs.values_list('f', flat=True)[0], [])
        self.assertTrue(qs.filter(f=[]).exists())
        self.assertTrue(qs.filter(f="").exists())
        TestModel.objects.all().delete()

        TestModel.objects.create(f="")
        self.assertEqual(qs.values_list('f', flat=True)[0], [])
        self.assertTrue(qs.filter(f=[]).exists())
        self.assertTrue(qs.filter(f="").exists())
        TestModel.objects.all().delete()

        TestModel.objects.create(f=None)
        self.assertEqual(TestModel.objects.values_list('f', flat=True)[0], [])

        self.assertTrue(qs.filter(f=None).exists())


class MultiEmailFormFieldTest(SimpleTestCase):

    def test__widget(self):
        f = MultiEmailFormField()
        self.assertIsInstance(f.widget, MultiEmailWidget)

    def test__to_python(self):
        f = MultiEmailFormField()
        # Empty values
        for val in ['', None]:
            self.assertEquals([], f.to_python(val))
        # One line correct value
        val = '  foo@bar.com    '
        self.assertEquals(['foo@bar.com'], f.to_python(val))
        # Multi lines correct values (test of #0010614)
        val = 'foo@bar.com\nfoo2@bar2.com\r\nfoo3@bar3.com'
        self.assertEquals(['foo@bar.com', 'foo2@bar2.com', 'foo3@bar3.com'],
                          f.to_python(val))

    def test__validate(self):
        f = MultiEmailFormField(required=True)
        # Empty value
        val = []
        self.assertRaises(ValidationError, f.validate, val)
        # Incorrect value
        val = ['not-an-email.com']
        self.assertRaises(ValidationError, f.validate, val)
        # An incorrect value with correct values
        val = ['foo@bar.com', 'not-an-email.com', 'foo3@bar3.com']
        self.assertRaises(ValidationError, f.validate, val)
        # Should not happen (to_python do the strip)
        val = ['  foo@bar.com    ']
        self.assertRaises(ValidationError, f.validate, val)
        # Correct value
        val = ['foo@bar.com']
        f.validate(val)


class MultiEmailWidgetTest(SimpleTestCase):

    def test__prep_value__empty(self):
        w = MultiEmailWidget()
        value = w.prep_value('')
        self.assertEqual(value, '')

    def test__prep_value__string(self):
        w = MultiEmailWidget()
        value = w.prep_value('foo@foo.fr\nbar@bar.fr')
        self.assertEqual(value, 'foo@foo.fr\nbar@bar.fr')

    def test__prep_value__list(self):
        w = MultiEmailWidget()
        value = w.prep_value(['foo@foo.fr', 'bar@bar.fr'])
        self.assertEqual(value, 'foo@foo.fr\nbar@bar.fr')

    def test__prep_value__raise(self):
        w = MultiEmailWidget()
        self.assertRaises(ValidationError, w.prep_value, 42)

    def test__render(self):
        w = MultiEmailWidget()
        output = w.render('test', ['foo@foo.fr', 'bar@bar.fr'])
        self.assertEqual(1, len(pq('textarea', output)))
        # The template-based widget add a line-return
        value = pq('textarea', output).text()
        self.assertEqual(
            value.lstrip(),
            'foo@foo.fr\nbar@bar.fr'
        )



from django.db import models

from multi_email_field.fields import MultiEmailField


class TestModel(models.Model):
    f = MultiEmailField(null=True, blank=True)
    f_default = MultiEmailField(default=["test@example.com"], null=True, blank=True)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

"""
Mixin class methods to be added to django.forms.fields at runtime. These methods add additional
error messages for AngularJS form validation.
"""
import re
from django.forms import fields
from django.forms import widgets
from django.utils.translation import gettext_lazy, ngettext_lazy
#from .widgets import CheckboxSelectMultiple as DjngCheckboxSelectMultiple


class DefaultFieldMixin(object):
    def get_potential_errors(self):
        return self.get_input_required_errors()

    def get_input_required_errors(self):
        errors = []
        if self.required:
            self.widget.attrs['ng-required'] = 'true'
            for key, msg in self.error_messages.items():
                if key == 'required':
                    errors.append(('$error.required', msg))
        return errors

    def get_min_max_length_errors(self):
        errors = []
        if getattr(self, 'min_length', None):
            self.widget.attrs['ng-minlength'] = self.min_length
        if getattr(self, 'max_length', None):
            self.widget.attrs['ng-maxlength'] = self.max_length
        for item in self.validators:
            if getattr(item, 'code', None) == 'min_length':
                message = ngettext_lazy(
                    'Ensure this value has at least %(limit_value)d character',
                    'Ensure this value has at least %(limit_value)d characters',
                    'limit_value')
                errors.append(('$error.minlength', message % {'limit_value': self.min_length}))
            if getattr(item, 'code', None) == 'max_length':
                message = ngettext_lazy(
                    'Ensure this value has at most %(limit_value)d character',
                    'Ensure this value has at most %(limit_value)d characters',
                    'limit_value')
                errors.append(('$error.maxlength', message % {'limit_value': self.max_length}))
        return errors

    def get_min_max_value_errors(self):
        errors = []
        if isinstance(getattr(self, 'min_value', None), int):
            self.widget.attrs['min'] = self.min_value
        if isinstance(getattr(self, 'max_value', None), int):
            self.widget.attrs['max'] = self.max_value
        errkeys = []
        for key, msg in self.error_messages.items():
            if key == 'min_value':
                errors.append(('$error.min', msg))
                errkeys.append(key)
            if key == 'max_value':
                errors.append(('$error.max', msg))
                errkeys.append(key)
        for item in self.validators:
            if getattr(item, 'code', None) == 'min_value' and 'min_value' not in errkeys:
                errors.append(('$error.min', item.message % {'limit_value': self.min_value}))
                errkeys.append('min_value')
            if getattr(item, 'code', None) == 'max_value' and 'max_value' not in errkeys:
                errors.append(('$error.max', item.message % {'limit_value': self.max_value}))
                errkeys.append('max_value')
        return errors

    def get_invalid_value_errors(self, ng_error_key):
        errors = []
        errkeys = []
        for key, msg in self.error_messages.items():
            if key == 'invalid':
                errors.append(('$error.{0}'.format(ng_error_key), msg))
                errkeys.append(key)
        for item in self.validators:
            if getattr(item, 'code', None) == 'invalid' and 'invalid' not in errkeys:
                errmsg = getattr(item, 'message', gettext_lazy('This input self does not contain valid data.'))
                errors.append(('$error.{0}'.format(ng_error_key), errmsg))
                errkeys.append('invalid')
        return errors


class CharFieldMixin(DefaultFieldMixin):
    def get_potential_errors(self):
        errors = self.get_input_required_errors()
        errors.extend(self.get_min_max_length_errors())
        return errors


class DecimalFieldMixin(DefaultFieldMixin):
    def get_potential_errors(self):
        errors = self.get_input_required_errors()
        self.widget.attrs['ng-minlength'] = 1
        if hasattr(self, 'max_digits') and self.max_digits > 0:
            self.widget.attrs['ng-maxlength'] = self.max_digits + 1
        errors.extend(self.get_min_max_value_errors())
        errors.extend(self.get_invalid_value_errors('number'))
        return errors


class EmailFieldMixin(DefaultFieldMixin):
    def get_potential_errors(self):
        self.widget.attrs['email-pattern'] = self.get_email_regex()
        errors = self.get_input_required_errors()
        errors.extend(self.get_invalid_value_errors('email'))
        return errors

    def get_email_regex(self):
        """
        Return a regex pattern matching valid email addresses. Uses the same
        logic as the django validator, with the folowing exceptions:

        - Internationalized domain names not supported
        - IP addresses not supported
        - Strips lookbehinds (not supported in javascript regular expressions)
        """
        validator = self.default_validators[0]
        user_regex = validator.user_regex.pattern.replace('\Z', '@')
        domain_patterns = ([re.escape(domain) + '$' for domain in
                            validator.domain_whitelist] +
                           [validator.domain_regex.pattern.replace('\Z', '$')])
        domain_regex = '({0})'.format('|'.join(domain_patterns))
        email_regex = user_regex + domain_regex
        return re.sub(r'\(\?\<[^()]*?\)', '', email_regex)  # Strip lookbehinds


class DateFieldMixin(DefaultFieldMixin):
    def get_potential_errors(self):
        errors = self.get_input_required_errors()
        errors.extend(self.get_invalid_value_errors('date'))
        return errors


class FloatFieldMixin(DefaultFieldMixin):
    def get_potential_errors(self):
        errors = self.get_input_required_errors()
        errors.extend(self.get_min_max_value_errors())
        errors.extend(self.get_invalid_value_errors('number'))
        return errors


class IntegerFieldMixin(DefaultFieldMixin):
    def get_potential_errors(self):
        errors = self.get_input_required_errors()
        errors.extend(self.get_min_max_value_errors())
        errors.extend(self.get_invalid_value_errors('number'))
        return errors


class SlugFieldMixin(DefaultFieldMixin):
    def get_potential_errors(self):
        errors = self.get_input_required_errors()
        return errors


class RegexFieldMixin(DefaultFieldMixin):
    # Presumably Python Regex can't be translated 1:1 into JS regex. Any hints on how to convert these?
    def get_potential_errors(self):
        self.widget.attrs['ng-pattern'] = '/{0}/'.format(self.regex.pattern)
        errors = self.get_input_required_errors()
        errors.extend(self.get_min_max_length_errors())
        errors.extend(self.get_invalid_value_errors('pattern'))
        return errors


class BooleanFieldMixin(DefaultFieldMixin):
    def get_potential_errors(self):
        errors = self.get_input_required_errors()
        return errors


class MultipleFieldMixin(DefaultFieldMixin):
    def get_multiple_choices_required(self):
        """
        Add only the required message, but no 'ng-required' attribute to the input fields,
        otherwise all Checkboxes of a MultipleChoiceField would require the property "checked".
        """
        errors = []
        if self.required:
            for key, msg in self.error_messages.items():
                if key == 'required':
                    errors.append(('$error.required', msg))
        return errors


class ChoiceFieldMixin(MultipleFieldMixin):
    def get_potential_errors(self):
        if isinstance(self.widget, widgets.RadioSelect):
            errors = self.get_multiple_choices_required()
        else:
            errors = self.get_input_required_errors()
        return errors


class MultipleChoiceFieldMixin(MultipleFieldMixin):
    def get_potential_errors(self):
        if isinstance(self.widget, widgets.CheckboxSelectMultiple):
            errors = self.get_multiple_choices_required()
        else:
            errors = self.get_input_required_errors()
        return errors

#    def get_converted_widget(self):
#        assert(isinstance(self, fields.MultipleChoiceField))
#        if not isinstance(self.widget, widgets.CheckboxSelectMultiple):
#            return
#        new_widget = DjngCheckboxSelectMultiple()
#        new_widget.__dict__ = self.widget.__dict__
#        return new_widget


class FileFieldMixin(DefaultFieldMixin):
    def get_potential_errors(self):
        errors = []
        return errors



from django.db import models

from multi_email_field.forms import MultiEmailField as MultiEmailFormField

try:
    from django.utils import six

    string_types = six.string_types
except ImportError:
    string_types = str


class MultiEmailField(models.Field):
    description = "A multi e-mail field stored as a multi-lines text"
    
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('default', [])
        super(MultiEmailField, self).__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        # This is a fairly standard way to set up some defaults
        # while letting the caller override them.
        defaults = {'form_class': MultiEmailFormField}
        defaults.update(kwargs)
        return super(MultiEmailField, self).formfield(**defaults)

    def from_db_value(self, value, expression, connection, context=None):
        if value is None:
            return []
        return value.splitlines()

    def get_db_prep_value(self, value, connection, prepared=False):
        if isinstance(value, string_types):
            return value
        elif isinstance(value, list):
            return "\n".join(value)

    def to_python(self, value):
        if not value:
            return []
        if isinstance(value, list):
            return value
        return value.splitlines()

    def get_internal_type(self):
        return 'TextField'



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json
from django.forms import widgets
from django.utils.safestring import mark_safe
from django.utils.encoding import force_str
from django.utils.html import format_html, format_html_join


def flatatt(attrs):
    """
    Pilfered from `django.forms.utils`:
    Convert a dictionary of attributes to a single string.
    The returned string will contain a leading space followed by key="value",
    XML-style pairs. In the case of a boolean value, the key will appear
    without a value. Otherwise, the value is formatted through its own dict of `attrs`,
    which can be useful to parametrize Angular directives.
    It is assumed that the keys do not need to be
    XML-escaped. If the passed dictionary is empty, then return an empty
    string.

    The result is passed through 'mark_safe' (by way of 'format_html_join').
    """
    key_value_attrs = []
    boolean_attrs = []
    for attr, value in attrs.items():
        if isinstance(value, bool):
            if value:
                boolean_attrs.append((attr,))
        else:
            try:
                value = value.format(**attrs)
            except KeyError:
                pass
            key_value_attrs.append((attr, value))

    return (
        format_html_join('', ' {}="{}"', sorted(key_value_attrs)) +
        format_html_join('', ' {}', sorted(boolean_attrs))
    )


#class ChoiceFieldRenderer(widgets.ChoiceFieldRenderer):
#    def render(self):
#        """
#        Outputs a <ul ng-form="name"> for this set of choice fields to nest an ngForm.
#        """
#        start_tag = format_html('<ul {0}>', mark_safe(' '.join(self.field_attrs)))
#        output = [start_tag]
#        for widget in self:
#            output.append(format_html('<li>{0}</li>', force_str(widget)))
#        output.append('</ul>')
#        return mark_safe('\n'.join(output))
#
#
#class CheckboxChoiceInput(widgets.CheckboxChoiceInput):
#    def tag(self, attrs=None):
#        attrs = attrs or self.attrs
#        name = '{0}.{1}'.format(self.name, self.choice_value)
#        tag_attrs = dict(attrs, type=self.input_type, name=name, value=self.choice_value)
#        if 'id' in attrs:
#            tag_attrs['id'] = '{0}_{1}'.format(attrs['id'], self.index)
#        if 'ng-model' in attrs:
#            tag_attrs['ng-model'] = "{0}['{1}']".format(attrs['ng-model'], self.choice_value)
#        if self.is_checked():
#            tag_attrs['checked'] = 'checked'
#        return format_html('<input{0} />', flatatt(tag_attrs))


class CheckboxFieldRendererMixin(object):
    def __init__(self, name, value, attrs, choices):
        attrs.pop('djng-error', None)
        self.field_attrs = [format_html('ng-form="{0}"', name)]
        if attrs.pop('multiple_checkbox_required', False):
            field_names = [format_html('{0}.{1}', name, choice) for choice, dummy in choices]
            self.field_attrs.append(format_html('validate-multiple-fields="{0}"', json.dumps(field_names)))
        super(CheckboxFieldRendererMixin, self).__init__(name, value, attrs, choices)


#class CheckboxFieldRenderer(CheckboxFieldRendererMixin, ChoiceFieldRenderer):
#    choice_input_class = CheckboxChoiceInput
#
#
#class CheckboxSelectMultiple(widgets.CheckboxSelectMultiple):
#    """
#    Form fields of type 'MultipleChoiceField' using the widget 'CheckboxSelectMultiple' must behave
#    slightly different from the original. This widget overrides the default functionality.
#    """
#    renderer = CheckboxFieldRenderer
#
#    def implode_multi_values(self, name, data):
#        """
#        Due to the way Angular organizes it model, when Form data is sent via a POST request,
#        then for this kind of widget, the posted data must to be converted into a format suitable
#        for Django's Form validation.
#        """
#        mkeys = [k for k in data.keys() if k.startswith(name + '.')]
#        mvls = [data.pop(k)[0] for k in mkeys]
#        if mvls:
#            data.setlist(name, mvls)
#
#    def convert_ajax_data(self, field_data):
#        """
#        Due to the way Angular organizes it model, when this Form data is sent using Ajax,
#        then for this kind of widget, the sent data has to be converted into a format suitable
#        for Django's Form validation.
#        """
#        return [key for key, val in field_data.items() if val]
#
#    def get_field_attrs(self, field):
#        return {'multiple_checkbox_required': field.required}


class RadioFieldRendererMixin(object):
    def __init__(self, name, value, attrs, choices):
        attrs.pop('djng-error', None)
        self.field_attrs = []
        if attrs.pop('radio_select_required', False):
            self.field_attrs.append(format_html('validate-multiple-fields="{0}"', name))
        super(RadioFieldRendererMixin, self).__init__(name, value, attrs, choices)


#class RadioFieldRenderer(RadioFieldRendererMixin, ChoiceFieldRenderer):
#    choice_input_class = widgets.RadioChoiceInput
#
#
#class RadioSelect(widgets.RadioSelect):
#    """
#    Form fields of type 'ChoiceField' using the widget 'RadioSelect' must behave
#    slightly different from the original. This widget overrides the default functionality.
#    """
#    renderer = RadioFieldRenderer
#
#    def get_field_attrs(self, field):
#        return {'radio_select_required': field.required}



# -*- coding: utf-8 -*-
#
# django-codenerix
#
# Codenerix GNU
#
# Project URL : http://www.codenerix.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

try:
    from subprocess import getstatusoutput
    pythoncmd = "python3"
except Exception:
    from commands import getstatusoutput
    pythoncmd = "python2"

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from codenerix.lib.debugger import Debugger


class Command(BaseCommand, Debugger):

    # Show this when the user types help
    help = "Remove *.pyc files"

    def handle(self, *args, **options):

        # Autoconfigure Debugger
        self.set_name("CODENERIX")
        self.set_debug()

        # Get environment
        appname = settings.ROOT_URLCONF.split(".")[0]
        basedir = settings.BASE_DIR
        appdir = os.path.abspath("{}/{}".format(basedir, appname))
        status, output = getstatusoutput("find {}/ -name '*.py[c|o]' -o -name __pycache__ -exec rm -rf {{}} +".format(appdir))
        if status:
            raise CommandError(output)



# -*- coding: utf-8 -*-
#
# django-codenerix
#
# Codenerix GNU
#
# Project URL : http://www.codenerix.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import pyotp
except ImportError:
    pyotp = None
import hashlib
import base64
import ssl
import datetime
import time

from codenerix_lib.debugger import Debugger

from django.utils import timezone
from django.contrib.auth import logout
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.models import User, Group
from django.contrib.auth import authenticate, login
from django.conf import settings

import ldap3
from ldap3.core.exceptions import LDAPException, LDAPSocketOpenError


def check_auth(user, debugger=None):
    """
    Check if the user should or shouldn't be inside the system:
    - If the user is staff or superuser: LOGIN GRANTED
    - If the user has a Person and it is not "disabled": LOGIN GRANTED
    - Elsewhere: LOGIN DENIED
    """

    # Initialize authentication
    auth = None
    person = None

    # Check if there is an user
    if user:

        # Show debugger
        if debugger:
            debugger("check_auth(): Username: '{}'".format(user), color="white")

        # It means that Django accepted the user and it is active
        if user.is_staff or user.is_superuser:

            # This is an administrator, let it in
            auth = user

            # Show debugger
            if debugger:
                debugger(
                    "check_auth(): user '{}' is an Adminitrator".format(user),
                    color="green",
                )

        else:

            # Show debugger
            if debugger:
                debugger(
                    "check_auth(): user '{}' is a Normal User".format(user),
                    color="cyan",
                )

            # It is a normal user, check if there is a person behind
            person = getattr(user, "person", None)
            if not person:

                # Show debugger
                if debugger:
                    debugger(
                        "check_auth(): no associated person found for user '{}'".format(
                            user
                        ),
                        color="yellow",
                    )

                # Check if there is related one
                person_related = getattr(user, "people", None)
                if person_related:

                    # Show debugger
                    if debugger:
                        debugger(
                            "check_auth(): there are People associated for user '{}'".format(
                                user
                            ),
                            color="cyan",
                        )

                    # Must be only one
                    if person_related.count() == 1:

                        # Person found
                        person = person_related.get()

                    elif debugger:
                        debugger(
                            "check_auth(): found more than One People associated for user '{}', not valid Person found!".format(
                                user
                            ),
                            color="yellow",
                        )

                elif debugger:
                    debugger(
                        "check_auth(): no People found for for user '{}', not valid Person found!".format(
                            user
                        ),
                        color="red",
                    )

            # Show debugger
            if debugger and person:
                debugger(
                    "check_auth(): user={} - person={} - disabled={}'".format(
                        user, person, person.disabled
                    ),
                    color="cyan",
                )

            # Check if the person is already disabled
            if person and (
                (person.disabled is None) or (person.disabled > timezone.now())
            ):
                # There is a person, no disabled found or the found one is fine to log in
                auth = user

                if debugger:
                    debugger(
                        "check_auth(): user '{}' access granted!".format(user),
                        color="green",
                    )

            elif debugger:
                debugger(
                    "check_auth(): user '{}' access NOT granted!".format(user),
                    color="red",
                )

    elif debugger:
        debugger(
            "check_auth(): No username, access NOT granted!".format(user), color="red"
        )

    # Return back the final decision
    return auth


class LimitedAuth(ModelBackend, Debugger):
    """
    Authentication system based on default Django's authentication system
    which extends the last one with check_auth() extra system limitations
    """

    def __init__(self, *args, **kwargs):

        # Configure debugger
        self.__debugger = getattr(settings, "AUTHENTICATION_DEBUG", False)
        if self.__debugger:
            self.set_debug()

        # Keep going with super
        super(LimitedAuth, self).__init__(*args, **kwargs)

    def authenticate(self, *args, **kwargs):

        # Show debugger
        if self.__debugger:
            self.debug("Started authenticate()", color="blue")

        # Launch default django authentication
        user = super(LimitedAuth, self).authenticate(*args, **kwargs)

        # Answer to the system
        answer = check_auth(user, self.debug)
        return answer


class LimitedAuthMiddleware(Debugger):
    """
    Check every request if the user should or shouldn't be inside the system
    SESSION_EXPIRE_WHEN_INACTIVE: number of seconds the user can be innactive on the website without being logged out
    SESSION_SHIFTS: list of hours of the day when all users must be logged out

    NOTE: install in your MIDDLEWARE setting after (order matters):
        'django.contrib.auth.middleware.AuthenticationMiddleware'
    """

    def __init__(self, get_response=None):

        # Configure debugger
        self.__debugger = getattr(settings, "AUTHENTICATION_DEBUG", False)
        if self.__debugger:
            self.set_debug()

        # Get response
        self.get_response = get_response

    def process_request(self, request):

        # Show debugger
        if self.__debugger:
            self.debug("Started process_request()", color="blue")

        # If the user is authenticated and shouldn't be
        if request.user.is_authenticated:

            # Show debugger
            if self.__debugger:
                self.debug("User already authenticated", color="cyan")

            # If the user doesn't pass the check_auth test
            if not check_auth(request.user, self.debug):

                # Show debugger
                if self.__debugger:
                    self.debug("User didn'pass check_auth()", color="red")

                # Push it out from the system
                logout(request)

            # Show debugger
            elif self.__debugger:
                self.debug("User pass check_auth()", color="green")

            # Get now
            now = datetime.datetime.now()

            # Create delta for caducity by EXPIRE_WHEN_INACTIVE
            expire_when_innactive = getattr(
                settings, "SESSION_EXPIRE_WHEN_INACTIVE", None
            )
            if expire_when_innactive is not None:
                # Get last_seen
                last_seen = request.session.get("user_last_seen", None)
                # Calculate delta
                if last_seen is not None:
                    # Check if we have to logout this user for being innactive too long
                    try:
                        # python3
                        nowts = now.timestamp()
                    except AttributeError:
                        # python2
                        nowts = float(
                            time.mktime(now.timetuple()) + now.microsecond / 1000000.0
                        )

                    diff = nowts - last_seen - expire_when_innactive

                    # Get caducity
                    if diff > 0:
                        # Push it out from the system
                        logout(request)

                # Refresh last_seen
                last_seen = datetime.datetime.now()
                try:
                    # python3
                    last_seen = last_seen.timestamp()
                except AttributeError:
                    # python2
                    last_seen = float(
                        time.mktime(last_seen.timetuple())
                        + last_seen.microsecond / 1000000.0
                    )

                # Remember in session
                request.session["user_last_seen"] = last_seen

            # Create delta for caducity by SHIFTS
            shifts = getattr(settings, "SESSION_SHIFTS", None)
            if shifts is not None:
                shifts = shifts + [shifts[0] + 24]
                for turno in shifts:
                    if turno > now.hour:
                        delta = turno % 24
                        break

                # Calculate caducity
                kickout = datetime.datetime(now.year, now.month, now.day, delta)
                try:
                    # python3
                    kickout = kickout.timestamp() - now.timestamp()
                except AttributeError:
                    # python2
                    ts = float(
                        time.mktime(kickout.timetuple())
                        + kickout.microsecond / 1000000.0
                    )
                    tz = float(
                        time.mktime(now.timetuple()) + now.microsecond / 1000000.0
                    )
                    kickout = ts - tz

                # Get caducity
                caducity = request.session.get("user_session_caducity", None)
                if not caducity or kickout < caducity:
                    # Remember new caducity
                    request.session.set_expiry(kickout)
                    request.session["user_session_caducity"] = kickout
                else:
                    # Push it out from the system
                    request.session.pop("user_session_caducity")
                    logout(request)

        elif self.__debugger:
            self.debug("User not authenticated", color="yellow")

    def __call__(self, request):

        # Code to be executed for each request before the view (and later middleware) are called.
        self.process_request(request)

        # Get response
        response = self.get_response(request)

        # Code to be executed for each request/response after the view is called
        # ... pass ...

        # Return response
        return response


class TokenAuth(ModelBackend, Debugger):
    """
    Authentication system based on a Token key
    """

    def __init__(self, *args, **kwargs):

        # Configure debugger
        self.__debugger = getattr(settings, "AUTHENTICATION_DEBUG", False)
        if self.__debugger:
            self.set_debug()

        # Keep going with super
        super(TokenAuth, self).__init__(*args, **kwargs)

    def authenticate(self, *args, **kwargs):

        # Show debugger
        if self.__debugger:
            self.debug("Started authenticate()", color="blue")

        # Get our arguments
        username = kwargs.get("username", None)
        token = kwargs.get("token", None)
        string = kwargs.get("string", "")

        # Show debug
        if self.__debugger:
            self.debug(
                "Authentication request:",
                color="white",
            )
            self.debug("  > Username [authuser]: '{}'".format(username), color="cyan")
            self.debug(
                "  > Token [authtoken]: '{}' <- authentication token (signed or unsigned)".format(
                    token
                ),
                color="cyan",
            )
            self.debug(
                "  > String [json]: '{}' <- string to be used as salt".format(string),
                color="cyan",
            )

        if username is not None and token is not None:

            # Try to find the user
            try:
                # Get the requested username
                user = User.objects.get(username=username)
            except User.DoesNotExist:
                user = None

            # If we got an user
            if user:

                # Show debug
                if self.__debugger:
                    self.debug(
                        "User '{}' found!".format(username),
                        color="green",
                    )

                # Get config
                config = {
                    "key": None,
                    "master_unsigned": False,
                    "user_unsigned": False,
                    "otp_unsigned": False,
                    "master_signed": False,
                    "user_signed": False,
                    "otp_signed": False,
                }
                config_settings = getattr(settings, "AUTHENTICATION_TOKEN", {})
                for (key, value) in config_settings.items():
                    config[key] = value

                # Show debug
                if self.__debugger:
                    self.debug(
                        "Authentication configuration:",
                        color="white",
                    )
                    for (key, value) in config_settings.items():
                        self.debug(
                            "  > {}: {}".format(key, value),
                            color="cyan",
                        )

                # Get keys
                if (
                    config["key"]
                    or config["master_unsigned"]
                    or config["master_signed"]
                ):
                    if config["key"] and (
                        config["master_unsigned"] or config["master_signed"]
                    ):

                        # Assign master key
                        master = config["key"]

                    else:
                        if self.__debugger:
                            raise IOError(
                                "To use a master key you have to set master_signed or master_unsigned to True and set 'master' to some valid string as your token"
                            )
                        else:
                            master = None
                else:
                    master = None

                if (
                    config["user_unsigned"]
                    or config["user_signed"]
                    or config["otp_unsigned"]
                    or config["otp_signed"]
                ):
                    # Check if user first_name is filled
                    if user.first_name:

                        # Get user first_name as user_key
                        user_key = user.first_name

                        if config["otp_unsigned"] or config["otp_signed"]:
                            if not pyotp:
                                raise IOError(
                                    "PYOTP library not found, you can not use OTP signed/unsigned configuration"
                                )
                            else:
                                try:
                                    otp = str(
                                        pyotp.TOTP(
                                            base64.b32encode(user_key.encode())
                                        ).now()
                                    )
                                except TypeError:
                                    if self.__debugger:
                                        raise IOError(
                                            "To use a OTP key you have to set a valid BASE32 string in the user's profile as your token, the string must be 16 characters long (first_name field in the user's model) - BASE32 string can have only this characters 'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567='. User {} key length is {} ".format(
                                                user_key, len(user_key)
                                            )
                                        )
                                    else:
                                        otp = None
                        else:
                            otp = None
                    else:
                        if self.__debugger:
                            raise IOError(
                                "To use a user/otp key you have to set user_signed, user_unsigned, otp_signed or otp_unsigned to True and set the user key in the user's profile to some valid string as your token (first_name field in the user's model)"
                            )
                        else:
                            user_key = None
                            otp = None
                else:
                    user_key = None
                    otp = None

                # Show debug
                if self.__debugger:
                    self.debug(
                        "Clear text keys:",
                        color="white",
                    )
                    if master:
                        info = []
                        if config.get("master_signed", False):
                            info.append("signed")
                        if config.get("master_unsigned", False):
                            info.append("unsigned")
                        self.debug(
                            "  > Master KEY configured: '{}' [{}]".format(
                                master, ", ".join(info)
                            ),
                            color="cyan",
                        )
                    else:
                        self.debug(
                            "  > Master KEY NOT configured",
                            color="purple",
                        )
                    if user_key:
                        info = []
                        if config.get("user_signed", False):
                            info.append("signed")
                        if config.get("user_unsigned", False):
                            info.append("unsigned")
                        self.debug(
                            "  > User KEY configured: '{}' [{}]".format(
                                user_key, ", ".join(info)
                            ),
                            color="cyan",
                        )
                    else:
                        self.debug(
                            "  > User KEY NOT configured",
                            color="purple",
                        )
                    if otp:
                        info = []
                        if config.get("otp_signed", False):
                            info.append("signed")
                        if config.get("otp_unsigned", False):
                            info.append("unsigned")
                        self.debug(
                            "  > OTP KEY configured: '{}' [{}]".format(
                                otp, ", ".join(info)
                            ),
                            color="cyan",
                        )
                    else:
                        self.debug(
                            "  > OTP KEY NOT configured",
                            color="purple",
                        )

                # Unsigned string
                if (
                    config["master_signed"]
                    or config["user_signed"]
                    or config["otp_signed"]
                ):
                    tosign = user.username + string

                # Show debugger
                if self.__debugger:
                    self.debug(
                        "Configured calculated keys:",
                        color="white",
                    )
                    self.debug(
                        "  > tosign = '{}' = '{}' + '{}'".format(
                            tosign, user.username, string
                        ),
                        color="white",
                    )

                # Build the list of valid keys
                keys = []
                if master:
                    if config["master_unsigned"]:  # MASTER KEY

                        # keys.append("master_unsigned")
                        keys.append(master)

                        # Show debugger
                        if self.__debugger:
                            self.debug(
                                "  > Master Unsigned Key: {}".format(keys[-1]),
                                color="cyan",
                            )

                    if config["master_signed"]:  # MASTER KEY SIGNED

                        # keys.append("master_signed")
                        keys.append(
                            hashlib.sha1(tosign.encode() + master.encode()).hexdigest()
                        )

                        # Show debugger
                        if self.__debugger:
                            self.debug(
                                "  > Master Signed Key: {} <- sha1( {} + {} )".format(
                                    keys[-1], tosign, master
                                ),
                                color="cyan",
                            )

                if user_key:
                    if config["user_unsigned"]:  # USER KEY

                        # keys.append("user_unsigned")
                        keys.append(user_key)

                        # Show debugger
                        if self.__debugger:
                            self.debug(
                                "  > User Unsigned Key: {}".format(keys[-1]),
                                color="cyan",
                            )

                    if config["user_signed"]:  # USER KEY SIGNED

                        # keys.append("user_signed")
                        keys.append(
                            hashlib.sha1(
                                tosign.encode() + user_key.encode()
                            ).hexdigest()
                        )

                        # Show debugger
                        if self.__debugger:
                            self.debug(
                                "  > User Signed Key: {} <- sha1( {} + {} )".format(
                                    keys[-1], tosign, user_key
                                ),
                                color="cyan",
                            )

                if otp:
                    if config["otp_unsigned"]:  # OTP KEY

                        # keys.append("otp_unsigned")
                        keys.append(otp)

                        # Show debugger
                        if self.__debugger:
                            self.debug(
                                "  > OTP Unsigned Key: {}".format(keys[-1]),
                                color="cyan",
                            )

                    if config["otp_signed"]:  # OTP KEY SIGNED

                        # keys.append("otp_signed")
                        keys.append(
                            hashlib.sha1(tosign.encode() + otp.encode()).hexdigest()
                        )

                        # Show debugger
                        if self.__debugger:
                            self.debug(
                                "  > OTP Signed Key: {} <- sha1( {}+ {} )".format(
                                    keys[-1], tosign, otp
                                ),
                                color="cyan",
                            )

                # Key is valid
                if token in keys:
                    answer = user

                    # Show debug
                    if self.__debugger:
                        self.debug(
                            "User '{}' authenticated!".format(username),
                            color="green",
                        )

                else:

                    # Not authenticated
                    answer = None

                    # Show debug
                    if self.__debugger:
                        self.debug(
                            "User '{}' NOT authenticated with tokenkey '{}'!".format(
                                username, token
                            ),
                            color="red",
                        )

            else:

                # Username not found, not accepting the authentication request
                answer = None

                # Show debug
                if self.__debugger:
                    self.debug(
                        "User '{}' NOT found!".format(username),
                        color="yellow",
                    )

        else:

            # Missing data, can not authenticate with this information
            answer = None

            # Show debug
            if self.__debugger:
                self.debug(
                    "Missing data, can not authenticte with this information",
                    color="yellow",
                )

        # Return answer
        return answer


class TokenAuthMiddleware(Debugger):
    """
    Check for every request if the user is not loged in, so we can log it in with a TOKEN

    NOTE 1: install in your MIDDLEWARE setting after (order matters):
        'django.contrib.auth.middleware.AuthenticationMiddleware'

    NOTE 2: if you are using POST with HTTPS, Django will require to send Referer, to avoid
        this problem you must add to the view of your url definition csrf_exempt(), as follows:

        from django.views.decorators.csrf import csrf_exempt
        urlpatterns = patterns(
            # ...
            # Will exclude `/api/v1/test` from CSRF
            url(r'^api/v1/test', csrf_exempt(TestApiHandler.as_view()))
            # ...
        )

        Check: http://stackoverflow.com/questions/11374382/how-can-i-disable-djangos-csrf-protection-only-in-certain-cases
        They recommend in this post to use the decorator, but we didn't manage to make it work
        in the post() method inside our class-view. Probably this will work in the dispatch().
    """

    def __init__(self, get_response=None):

        # Configure debugger
        self.__debugger = getattr(settings, "AUTHENTICATION_DEBUG", False)
        if self.__debugger:
            self.set_debug()

        # Get response
        self.get_response = get_response

    def process_request(self, request):

        # Show debugger
        if self.__debugger:
            self.debug("Started process_request()", color="blue")

        # By default we are not in authtoken
        request.authtoken = False

        # Get body
        body = request.body

        # Get token
        token = request.GET.get("authtoken", request.POST.get("authtoken", ""))

        # If the user is authenticated and shouldn't be
        if token:

            if self.__debugger:
                self.debug("AUTHTOKEN='{}'".format(token), color="cyan")

            # Get username and json
            username = request.GET.get("authuser", request.POST.get("authuser", None))
            json = request.GET.get("json", request.POST.get("json", body.decode()))
            if self.__debugger:
                self.debug("USERNAME={} - JSON={}".format(username, json), color="cyan")

            # Authenticate user
            user = authenticate(
                request=None, username=username, token=token, string=json
            )
            if user:

                # Show debug
                if self.__debugger:
                    self.debug("User authenticated", color="green")

                # Set we are in authtoken
                request.authtoken = True
                # Log user in
                login(request, user)
                # Disable CSRF checks
                setattr(request, "_dont_enforce_csrf_checks", True)
                json_details = request.GET.get(
                    "authjson_details", request.POST.get("authjson_details", False)
                )
                if json_details in ["true", "1", "t", True]:
                    json_details = True
                else:
                    json_details = False
                request.json_details = json_details

            elif self.__debugger:
                self.debug("User not authenticated", color="red")

        elif self.__debugger:
            self.debug("No authtoken found in your request", color="yellow")

    def __call__(self, request):

        # Code to be executed for each request before the view (and later middleware) are called.
        self.process_request(request)

        # Get response
        response = self.get_response(request)

        # Code to be executed for each request/response after the view is called
        # ... pass ...

        # Return response
        return response


class ActiveDirectoryGroupMembershipSSLBackend:
    """
    Authorization backend for Active Directory in Django

    # Possible configuration parameters
    # AD_SSL = True                             # Use SSL
    # AD_CERT_FILE='/path/to/your/cert.txt'     # Path to SSL certificate
    # AD_DEBUG_FILE='/tmp/ldap.debug'           # Path to DEBUG file (if none, Debugging will be disabled)
    # AD_LDAP_PORT=9834                         # Port to use
    # AD_DNS_NAME='NTDOMAIN.CODENERIX.COM'      # DNS nameserver if different thatn NT4 DOMAIN
    AD_LOCK_UNAUTHORIZED=True                   # Unauthorized users in Active Directory should be locked in Django
    AD_NT4_DOMAIN='NTDOMAIN.CODENERIX.COM'      # NT4 Domain name
    AD_MAP_FIELDS= {                            # Fields to map:   left=Django   right=Active Directory
            'email':        'mail',
            'first_name':   'givenName',
            'last_name':    'sn',
        }
    """

    __debug = None

    def debug(self, msg):
        """
        Handle the debugging to a file
        """
        # If debug is not disabled
        if self.__debug is not False:

            # If never was set, try to set it up
            if self.__debug is None:

                # Check what do we have inside settings
                debug_filename = getattr(settings, "AD_DEBUG_FILE", None)
                if debug_filename:
                    # Open the debug file pointer
                    self.__debug = open(settings.AD_DEBUG_FILE, "a")
                else:
                    # Disable debuging forever
                    self.__debug = False

            if self.__debug:
                # Debug the given message
                self.__debug.write("{}\n".format(msg))
                self.__debug.flush()

    def ldap_link(self, username, password, mode="LOGIN"):

        # If no password provided, we will not try to authenticate
        if password:

            # Prepare LDAP connection details
            nt4_domain = settings.AD_NT4_DOMAIN.upper()
            dns_name = getattr(settings, "AD_DNS_NAME", nt4_domain).upper()
            use_ssl = getattr(settings, "AD_SSL", False)
            if use_ssl:
                default_port = 636
                proto = "ldaps"
            else:
                default_port = 389
                proto = "ldap"
            port = getattr(settings, "AD_LDAP_PORT", default_port)
            ldap_url = "{}://{}:{}".format(proto, dns_name, port)
            self.debug("ldap.initialize :: url: {}".format(ldap_url))

            # Prepare library
            ser = {}
            ser["allowed_referral_hosts"] = [("*", True)]
            con = {}
            con["user"] = "{}\{}".format(nt4_domain, username)
            con["password"] = password
            con["raise_exceptions"] = True
            con["authentication"] = ldap3.NTLM
            if use_ssl:
                certfile = settings.AD_CERT_FILE
                self.debug("ldap.ssl :: Activated - Cert file: {}".format(certfile))
                con["auto_bind"] = ldap3.AUTO_BIND_TLS_BEFORE_BIND
                ser["use_ssl"] = True
                ser["tls"] = ldap3.Tls(
                    validate=ssl.CERT_REQUIRED, version=ssl.PROTOCOL_TLSv1
                )
            else:
                con["auto_bind"] = ldap3.AUTO_BIND_NO_TLS
            try:
                # self.debug('ldap.server params :: {}'.format(ser))
                server = ldap3.Server(ldap_url, **ser)
                self.debug("ldap.server :: {}".format(server))
                # self.debug('ldap.connection params :: {}'.format(con))
                answer = ldap3.Connection(server, **con)
                self.debug("ldap.connection :: {}".format(answer))
                # answer.open()
                # answer.bind()
                self.debug("ldap.connected :: Authorized")
            except LDAPSocketOpenError as e:
                # The access for this user has been denied, Debug it if required
                self.debug(
                    "LDAP connect failed 'SocketOpenError' for url '{}' with error '{}'".format(
                        ldap_url, e
                    )
                )
                answer = None
            except LDAPException as e:
                # The access for this user has been denied, Debug it if required
                self.debug(
                    "LDAP connect failed 'LDAPException' for user '{}' with error '{}'".format(
                        username, e
                    )
                )
                answer = False

        else:
            # The access for this user has been denied, Debug it if required
            self.debug("No password provided for user '{}'".format(username))
            answer = False

        # Return the final result
        return answer

    def authenticate(self, *args, **kwargs):
        """
        Authenticate the user agains LDAP
        """

        # Get config
        username = kwargs.get("username", None)
        password = kwargs.get("password", None)

        # Check user in Active Directory (authorization == None if can not connect to Active Directory Server)
        authorization = self.ldap_link(username, password, mode="LOGIN")

        if authorization:
            # The user was validated in Active Directory
            user = self.get_or_create_user(username, password)
            # Get or get_create_user will revalidate the new user
            if user:
                # If the user has been properly validated
                user.is_active = True
                user.save()
        else:
            # Locate user in our system
            user = User.objects.filter(username=username).first()
            if user and not user.is_staff:
                # If access was denied
                if authorization is False or getattr(
                    settings, "AD_LOCK_UNAUTHORIZED", False
                ):
                    # Deactivate the user
                    user.is_active = False
                    user.save()

            # No access and no user here
            user = None

        # Return the final decision
        return user

    def get_ad_info(self, username, password):

        self.debug("get_ad_info for user '{}'".format(username))

        # Initialize the answer
        info = {}

        # Get an already authenticated link connection to LDAP
        link = self.ldap_link(username, password, mode="SEARCH")

        if link:

            # Prepare SEARCH fields
            mapping = getattr(settings, "AD_MAP_FIELDS", {})
            # Build the search fields
            search_fields = ["sAMAccountName", "memberOf"] + list(mapping.values())
            search_dns = getattr(settings, "AD_NT4_DOMAIN", "").lower().split(".")
            # Build the dn list
            search_dnlist = []
            for token in search_dns:
                search_dnlist.append("dc={}".format(token))
            search_dn = ",".join(search_dnlist)

            # Search for the user
            self.debug('Search "{}"'.format(search_dn))
            # Search in LDAP
            link.search(
                search_base=search_dn,
                # search_filter='(&(objectclass=person)(uid=admin))',
                search_filter="(sAMAccountName={})".format(username),
                search_scope=ldap3.SUBTREE,
                # attributes=ldap3.ALL_ATTRIBUTES,
                attributes=search_fields,
                get_operational_attributes=True,
                size_limit=1,
            )

            # Get all results
            results = link.entries

            # Make sure we found only one result
            if len(results) == 1:

                # Get answer
                result = results[0].__dict__

            elif len(results) > 1:
                # Found serveral results
                self.debug("I found several results for your LDAP query")
                memberships = []
            else:
                # Not found
                self.debug("I didn't find any matching result for your LDAP query")
                memberships = []

            # Validate that they are a member of review board group
            memberships = result.get("memberOf", [])

            # Process all memberships found
            groupsAD = {}
            for membership in memberships:
                tokens = membership.split(",")
                dcs = []
                cn = None

                for token in tokens:
                    (key, value) = token.split("=")
                    if key == "CN":
                        if not cn:
                            cn = value
                    elif key == "DC":
                        dcs.append(value)

                # Prepare the full domain name addess of the AD
                dc = ".".join(dcs)

                # Make sure the key exists
                if dc not in groupsAD:
                    groupsAD[dc] = []

                # Add the new CN to the list
                if cn not in groupsAD[dc]:
                    groupsAD[dc].append(cn)

            # Prepare the answer
            info = {}
            info["groups"] = groupsAD

            # Look for other tokens to get mapped
            for djfield in mapping.keys():
                adfield = mapping[djfield]
                if adfield in result:
                    info[djfield] = result[adfield]
                    self.debug("{}={}".format(adfield, info[djfield]))
                else:
                    self.debug("{}=-NONE-".format(adfield))

        else:
            # No link gotten
            self.debug("I didn't get a valid link to the LDAP server")

        # Return the final result
        return info

    def get_or_create_user(self, username, password):
        """
        Get or create the given user
        """

        # Get the groups for this user
        info = self.get_ad_info(username, password)
        self.debug("INFO found: {}".format(info))

        # Find the user
        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist:
            user = User(username=username)

        # Update user
        user.first_name = info.get("first_name", "")
        user.last_name = info.get("last_name", "")
        user.email = info.get("email", "")

        # Check if the user is in the Administrators groups
        is_admin = False
        for domain in info["groups"]:
            if "Domain Admins" in info["groups"][domain]:
                is_admin = True
                break

        # Set the user permissions
        user.is_staff = is_admin
        user.is_superuser = is_admin

        # Refresh the password
        user.set_password(password)

        # Validate the selected user and gotten information
        user = self.validate(user, info)
        if user:
            self.debug("User got validated!")

            # Autosave the user until this point
            user.save()

            # Synchronize user
            self.synchronize(user, info)
        else:
            self.debug("User didn't pass validation!")

        # Finally return user
        return user

    def get_user(self, user_id):
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None

    def validate(self, user, info):
        """
        Dummy validate system, to be redeclared
        User object here is not saved in the database, but it is ready to be saved
        If the answer from this method is None, user will be denied to login in to the system!
        """
        self.debug("Validation process!")
        return user

    def synchronize(self, user, info):
        """
        It tries to do a group synchronization if possible
        This methods should be redeclared by the developer
        """

        self.debug("Synchronize!")

        # Remove all groups from this user
        user.groups.clear()

        # For all domains found for this user
        for domain in info["groups"]:
            # For all groups he is
            for groupname in info["groups"][domain]:
                # Lookup for that group
                group = Group.objects.filter(name=groupname).first()
                if group:
                    # If found, add the user to that group
                    user.groups.add(group)
